// BatteryStatus.cxx
//
// Copyright (C) 2000 Agenda Computing Inc.
// (C) 2001-2004 Brian Webb, Andreas Holzer, Enrique Vidal
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software Foundation,
// Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <FL/Fl.H>
#include <FL/Fl_Widget.H>
#include <FL/fl_draw.H>
#include <FL/filename.H>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <math.h>
#include <errno.h>
#include <stdlib.h>
#include "BatteryStatus.h"

int use_grey=0; // black

// New discharge tables borrowed from "flbatt"
int Alka[21] = {2000, 2000, 2000, 2052, 2105, 2147, 2190,
                2215, 2240, 2260, 2280, 2315, 2350, 2390,
                2430, 2485, 2540, 2615, 2690, 2820, 2950};

int NiCd[21] = {2200, 2310, 2355, 2392, 2412, 2433, 2449,
                2462, 2472, 2483, 2491, 2498, 2503, 2505,
                2508, 2510, 2511, 2513, 2517, 2533, 2600};

int NiMh[21] = {2092, 2339, 2421, 2468, 2506, 2531, 2554,
                2571, 2585, 2595, 2604, 2608, 2614, 2618,
                2619, 2624, 2637, 2666, 2721, 2799, 3043};

static int * curve=NiMh;

static int battlvl() {
#ifdef FAKE_BATTERY
  // Fake Battery bar bar showing ~50% of voltage range (0-3.3V):
  // return ADvalue * 100 / 1024;  Used for host compiling.
  return 512 * 100 / 1024;
#endif

#define n 10 // Number of batt voltage readings, for averaging. 
// Also aimed at doing some CPU work to enforce some batt current drain.
  int fd = open("/dev/adif", O_RDONLY); // Open the A-D converter device
  if(fd == -1) {perror("Opening /dev/adif"); return 0;}
  int avgVal=0, m=0;
  for (int i=0; i<n; i++) { // Perform n readings and average them
    /* read all 4 registers even if we only care about the first one */
    unsigned short value[4];
    if(read(fd, &value, 4*sizeof(short)) != 4 * sizeof(short)) {
      perror("Reading in /dev/adif"); return 0;
    }
    unsigned short v=value[0];
    if ((v>512) && (v<1024)) {avgVal += v; m++;}
//fprintf(stderr,"%d ", v);
  }
  close(fd);
  avgVal /= m;
//fprintf(stderr,"avgVal=%d -- ", avgVal);

  int mv =  (3300*avgVal) / 1024,  pc = 0, i, j;
  if      (mv < curve[0])  pc =   0;
  else if (mv > curve[20]) pc = 100;
  else {  // Find milivolts (mv) in curve[] and do liner interpolation
    for (i=0; i<21; i++) {if (mv >= curve[i]) j = i; else break;}
    pc = 5*j + (5*(mv-curve[j])) / (curve[j+1]-curve[j]);
  }
//fprintf(stderr,"mv=%d pc=%d\n", mv, pc);
  return(pc);
}

BatteryStatus::BatteryStatus (int x, int y, int w, int h, int type, int percent)
  : Fl_Button(x, y, w, h, "") {
  box(FL_FLAT_BOX);
  use_label = false;
  use_grey = 0;  // black
  if (type == 0) curve = Alka;
  if (type == 1) curve = NiCd;
  if (type == 2) curve = NiMh;
  if(percent == 0) level(battlvl()); else level(percent);
//fprintf(stderr,"battType=%d", type);
}

void BatteryStatus::level(int percent) {
  int p = percent;
  if(percent < 0) p = 0; if(percent > 100) p = 100;
  this->level_percent = p;
  if(use_label)
    {snprintf(label_buffer, 4, "%2d%%", p); label(label_buffer);}
  redraw();
}
int BatteryStatus::level(void) {
  return level_percent;
}

/////////////////////////////////////////////////////////////////////////
// Delay the real read of /dev/adif to avoid blocking other activities in 
// the strange cases when /dev/adif fails to supply data (typically after 
// long hibernation).  Also aims to give visual info of a blocked status.
void blinkUpdBatt(void* data) {
  BatteryStatus* v=(BatteryStatus*)data;
  static int m=0, grey=0; 
  if (m<11) {
    use_grey=grey; v->draw_box(); v->redraw(); Fl::flush();
    if (grey==0) grey=200; else grey=0;
    m++; Fl::repeat_timeout((unsigned int)200, blinkUpdBatt, data);}
  else {
    m=0; use_grey=110;  
    v->draw_box(); v->show();  v->redraw();  Fl::flush();
    int bl=battlvl(); 
    use_grey=0;  v->level(bl); v->draw_box();  
    v->show();   v->redraw();  Fl::flush();
//  fprintf(stderr, "--> batt=%d\n", bl);
    Fl::remove_timeout(blinkUpdBatt, v);
  }
}

void BatteryStatus::set() {
//int lev=battlvl(); level(lev); draw_box(); redraw(); // old code
  blinkUpdBatt(this);
}


//// Andreas' code used for experimenting with jffs2 
////
//// /* size of directory */
//// int getDirUsage(long *fa){
//// 
////   FILE *output;
//// 	char dummy[255] = "",
//// 	command[255],
//// 	*p1;
//// 	int i, pc;
//// 
////         filename_expand(command,"~/.wmx/.launchpad-flash-du");
//// 
////   output = popen (command, "r");
//// 
////   if (!output) {
////       fprintf (stderr, "Could not run flash du.\n");
////       return 0;
////   };
//// 
////   i=0;
////   do {
//// 
////      p1 = fgets(dummy, 255, output);
//// 
////      if(p1==NULL) break;
//// 
////      errno=0;
////      pc = strtol(p1,NULL,0);
////      i++;
////      if(errno == 0) {
////        *fa = pc;
////        fa++;
////      } else break;
////   } while (!feof(output));
//// 
////   pclose(output);
////   errno = 0;
////   return i;
//// }

void BatteryStatus::draw_box() {
  Fl_Widget::draw_box();
  fl_color(FL_BLACK);

  static int first=1, X=x(), Y=y(), W=w(), H=h(),
	     X1=X+1, X2=X+2, X4=X+4, H34=3*H/4, H2=H/2, H4=H/4;

  W -= 2;
////  W-=14;

  if (first) {
    fl_line(X,   Y, X2, Y);
    fl_line(X, H34, X1, H34);
    fl_line(X,  H2, X2, H2);
    fl_line(X,  H4, X1, H4);
    fl_line(X,   0, X2, 0);
  }

//printf("%d\n",level_percent);

  fl_line(X4, Y+H-1, X4+W-5, Y+H-1);     // _
  fl_line(X4, Y+2, X4,   Y+H-1);         //|_
  fl_line(X4+W-5, Y+2, X4+W-5,   Y+H-1); // |_|
  fl_line(X4, Y+2, X4+1,   Y+2);
  fl_line(X4+W-5, Y+2, X4+W-6,   Y+2);
  fl_line(X4+1, Y+2, X4+1,   Y+1);
  fl_line(X4+W-6, Y+2, X4+W-6,   Y+1);
  fl_line(X4+1, Y+1, X4+W-6,   Y+1);

  // for visual feedback of /dev/adif read blocking condition (adif.c bug)
  fl_color(use_grey,use_grey,use_grey);

  // make the solid rectangle strictly inside the outline
  fl_rectf(X4+1,  Y + H - (H-2)*level_percent / 100,
	   W-6,           (H-2)*level_percent / 100 - 1);
//fl_rectf(X4, Y + H -1 - (H-2)*level_percent / 100,  
//         W-5,           (H-2)*level_percent / 100);

//// Andreas' code used for experimenting with jffs2 
////
////   {
////      int i,il;
////      long fa[10];
//// 
////      il = getDirUsage(fa);
//// 
////      for(i=0; i<il; i++) {
////          fl_rectf(X4+W-3+i*4,  Y + H - (H-2)*fa[i] / 100,
//// 	          3,                   (H-2)*fa[i] / 100);
////      }
////  }
////
////  W += 14;

  W += 2;
}

void BatteryStatus::draw(void) {
  draw_box();
  draw_label();
}
void BatteryStatus::draw_label(void) {
  if(use_label)
    Fl_Widget::draw_label();
}
void BatteryStatus::with_label(void) {
  use_label = true;
  redraw();
}
void BatteryStatus::without_label(void) {
  use_label = false;
  redraw();
}
