// LaunchBrowser.cxx
//
// Copyright (C) 2000 Agenda Computing Inc.
// Written by Derek K.T. Au <phantom@netcom.ca>
// (C) 2001-2004 Brian Webb, Andreas Holzer, Enrique Vidal
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <algorithm>

#include <FL/Fl.H>
#include <FL/filename.H>
#include <FL/Fl_Box.h>

#include <flpda/fe.h>
#include <runone.h>

#include "Fl_Icon.h"
#include "Fl_XPM.h"
#include "Folder.h"
#include "LaunchBrowser.h"
#include "on_screen.h"
#include "folder_xpm.xpm"
#include "active_xpm.xpm"
#include "back_xpm.xpm"
#include "XActiveWindow.h"
#include "flpad.h"

using std::string;
using std::vector;

extern Fl_Box* launchpad_label;
// extern Fl_Button *back_button;
// extern Fl_Button *home_button;
Folder *LB_Action::folder = 0;

/*****************************************************************************
 * Utility functions.
 *****************************************************************************/

bool str_cmp_nocase(const std::string& a, const std::string& b) {
  int ret = strcasecmp(a.c_str(), b.c_str());
  if(ret < 0) {
    return true;
  }
  return false;
};


/*****************************************************************************
 * Launch Program
 *****************************************************************************/

class Launch_LB_Action : public LB_Action {
public:
  Launch_LB_Action(LaunchBrowser *lb, const string &prog);

  virtual LB_Action *copy() const {
    return new Launch_LB_Action(_lb, _prog);
  }

  virtual Fl_Pixmap *large_pixmap() const;

  virtual void action();

private:
  string _prog;
};

Launch_LB_Action::Launch_LB_Action(LaunchBrowser *lb, const string &prog)
  : LB_Action(lb), _prog(prog) {
  string::size_type slash_loc = _prog.rfind('/');
  if(slash_loc == string::npos)
    _label = _prog;
  else
    _label = string(_prog, slash_loc + 1, _label.size() - slash_loc);
}

Fl_Pixmap *Launch_LB_Action::large_pixmap() const {
  string::size_type slash_loc = _prog.rfind('/');
  string category = (slash_loc == string::npos) ? string("") :
    string(_prog, 0, slash_loc);
  return LB_Action::folder->get_pixmap(category.c_str(), _label.c_str());
}

void Launch_LB_Action::action() {
  on_screen(_label.c_str());
  string full_path = _lb->base_path + string("/") + _prog;
  char *argv[] = { (char*)full_path.c_str(), 0 };
  runone(argv);
}


/*****************************************************************************
 * Active Window
 *****************************************************************************/

class Window_LB_Action : public LB_Action {
public:
  Window_LB_Action(LaunchBrowser *lb, const string &name, Window xwin)
    : LB_Action(lb, name), _xwin(xwin) { }

  virtual LB_Action *copy() const {
    return new Window_LB_Action(_lb, _label, _xwin);
  }

  virtual Fl_Pixmap *large_pixmap() const;

  virtual void action() {
    XRaiseWindow(fl_display, _xwin);
  }

private:
  Window _xwin;
};

Fl_Pixmap *Window_LB_Action::large_pixmap() const {
  return LB_Action::folder->get_pixmap("Active", _label.c_str());
}

bool operator<(const Window_LB_Action &v1, const Window_LB_Action &v2) {
  return v1.label() < v2.label();
}


/*****************************************************************************
 * Back
 *****************************************************************************/

class Back_LB_Action : public LB_Action {
public:
  Back_LB_Action(LaunchBrowser *lb) : LB_Action(lb, string("Back")) {}

  virtual LB_Action *copy() const {
    return new Back_LB_Action(_lb);
  }

  virtual Fl_Pixmap *large_pixmap() const;

  virtual void action() {
    _lb->back();
  }
};

Fl_Pixmap *Back_LB_Action::large_pixmap() const {
  return Fl_XPM::copy(back_xpm);
}


/*****************************************************************************
 * Active Window List
 *****************************************************************************/

class WinList_LB_Action : public LB_Action {
public:
  WinList_LB_Action(LaunchBrowser *lb, const string &cat = string("Active"));

  virtual LB_Action *copy() const {
    return new WinList_LB_Action(_lb, _label);
  }

  virtual Fl_Pixmap *large_pixmap() const;

  virtual void action();

private:
  string _prog;
};

WinList_LB_Action::WinList_LB_Action(LaunchBrowser *lb, const string &cat)
  : LB_Action(lb, cat) { }

void WinList_LB_Action::action() {

  /*
   * When we delete the buttons we get deleted as well, so we do a little
   * trick here.  We need to create a copy of ourself to put on the action
   * stack anyway, so make that copy and call the action method of the copy,
   * which will not get deleted.  We check to see if we're the copy by
   * looking at the top of the action stack.
   */
  if(this != _lb->action_stack.top()) {
    _lb->action_stack.push(copy());
    _lb->action_stack.top()->action();
    return;
  }

  /*
   * Delete the buttons.
   */
  _lb->delete_buttons();

  /*
   * Add the back icon.
   */
  _lb->add_button(new Back_LB_Action(_lb));

  /*
   * Recreate the list of active windows.
   */
  vector<Window_LB_Action> active_list;
  for(XActiveWindowGen xagen(fl_display, fl_screen); !xagen.done(); ++xagen) {
    XActiveWindow aw = *xagen;
    if(aw.name != std::string("Launch Pad"))
      active_list.push_back(Window_LB_Action(_lb, aw.name, aw.xwindowid));
  }

  /*
   * Sort the list by name.
   */
  sort(active_list.begin(), active_list.end());

  /*
   * Create the buttons.
   */
  for(unsigned int i = 0; i < active_list.size(); i++)
    _lb->add_button(active_list[i].copy());

  /*
   * Set the current category to be the current directory name.
   */
  _lb->category(_label);
  _lb->redraw();
}

Fl_Pixmap *WinList_LB_Action::large_pixmap() const {
  return Fl_XPM::copy(active_xpm);
}


Fl_Pixmap *LB_Action::large_pixmap() const {
  return Fl_XPM::read(_lgpix_fname.c_str());
}

Fl_Pixmap *LB_Action::small_pixmap() const {
  return Fl_XPM::read(_smpix_fname.c_str());
}


/*****************************************************************************
 * Folder
 *****************************************************************************/

class Folder_LB_Action : public LB_Action {
public:
  Folder_LB_Action(LaunchBrowser *lb, const string &dir);

  virtual LB_Action *copy() const {
    return new Folder_LB_Action(_lb, _dir);
  }

  virtual void action();

  virtual Fl_Pixmap *large_pixmap() const;

private:
  string _dir;
};

Folder_LB_Action::Folder_LB_Action(LaunchBrowser *lb, const string &dir)
  : LB_Action(lb), _dir(dir) {
  string::size_type slash_loc = _dir.rfind('/');
  if(_dir == string(""))
    _label = string(ROOTPAD_NAME);
  else if(slash_loc == string::npos)
    _label = _dir;
  else
    _label = string(_dir, slash_loc, _dir.size() - slash_loc);
}

void Folder_LB_Action::action() {

  /*
   * When we delete the buttons we get deleted as well, so we do a little
   * trick here.  We need to create a copy of ourself to put on the action
   * stack anyway, so make that copy and call the action method of the copy,
   * which will not get deleted.  We check to see if we're the copy by
   * looking at the top of the action stack.
   */
  if(this != _lb->action_stack.top()) {
    _lb->action_stack.push(copy());
    _lb->action_stack.top()->action();
    return;
  }

  /*
   * Delete the buttons.
   */
  _lb->delete_buttons();

  /*
   * Open the directory with the full path name.
   */
  string full_path = _lb->base_path + string("/") + _dir;
  DIR* dir = opendir(full_path.c_str());
  if(!dir)
    return;

  /*
   * Parse through the directory.
   */
  vector<string> dir_list;
  vector<string> file_list;
  dirent* de;	
  while((de = readdir(dir)) != 0) {
    string fname(de->d_name);
    string sub_path = (_dir.size() == 0) ? fname : _dir + string("/") + fname;
    string name = _lb->base_path + string("/") + sub_path;

    struct stat sb;
    if(stat(name.c_str(), &sb) == 0)
      if(fname[0] != '.')
	if(S_ISDIR(sb.st_mode))
	  dir_list.push_back(string(sub_path));
	else if(fname != string("LaunchPad"))
	  file_list.push_back(string(sub_path));
  }
  closedir(dir);

  /*
   * Create the buttons.
   */
  if(_label == string(ROOTPAD_NAME))
    _lb->add_button(new WinList_LB_Action(_lb));
  else
    _lb->add_button(new Back_LB_Action(_lb));
  sort(dir_list.begin(), dir_list.end(), str_cmp_nocase);
  for(vector<string>::iterator i = dir_list.begin();
      i != dir_list.end(); i++)
    _lb->add_button(new Folder_LB_Action(_lb, *i));
  sort(file_list.begin(), file_list.end(), str_cmp_nocase);
  for(vector<string>::iterator i = file_list.begin();
      i != file_list.end(); i++)
    _lb->add_button(new Launch_LB_Action(_lb, *i));

  /*
   * Set the current category to be the current directory name.
   */
  _lb->category(_label);
  _lb->redraw();
}

Fl_Pixmap *Folder_LB_Action::large_pixmap() const {

    Fl_Pixmap *p = NULL;

    p =  LB_Action::folder->get_pixmap(
      ".",
      ("f-" + _label).c_str());

    if(p != NULL) return p;

    return Fl_XPM::copy(folder_xpm);
}


/*****************************************************************************
 * LaunchBrowser
 *****************************************************************************/

LaunchBrowser::LaunchBrowser(int x, int y, int w, int h)
  : Fl_Paged_Scroll(x, y, w ,h), sound(false) {
  char path[1024];
  filename_expand(path, "~/.wmx");
  base_path = string(path);

  folder = new Folder(base_path.c_str());
  LB_Action::folder = folder;

  current_folder = ROOTPAD_NAME;

  action_stack.push(new Folder_LB_Action(this, ""));

  refreshing = false;
}

LaunchBrowser::~LaunchBrowser() {
  delete_buttons();
  while(!action_stack.empty())
    delete action_stack.top();
}

void LaunchBrowser::home() {
  do_sound();
  while(action_stack.size() > 1) {
    delete action_stack.top();
    action_stack.pop();
  }
  refresh();
}

 void LaunchBrowser::icons() {
   folder->refresh_icon();
   refresh();
 }

void LaunchBrowser::back() {
  do_sound();
  if(action_stack.size() > 1) {
    delete action_stack.top();
    action_stack.pop();
  }
  refresh();
}

void LaunchBrowser::refresh() {
  refreshing = true;
  action_stack.top()->action();
  refreshing = false;
}

void LaunchBrowser::delete_buttons() {

  /*
   * Delete all the icons.
   */
  for(vector<Fl_Icon*>::iterator i = icon_list.begin();
      i != icon_list.end(); i++) {
    remove(*i);
    free((char*)(*i)->label());
    delete *i;
  }
  icon_list.clear();

  /*
   * Delete all the pixmaps.
   */
  for(vector<Fl_Pixmap*>::iterator i = pixmap_list.begin();
      i != pixmap_list.end(); i++)
    Fl_XPM::delete_pixmap(*i);
  pixmap_list.clear();

  /*
   * Delete all the actions.
   */
  for(vector<LB_Action*>::iterator i = callback_list.begin();
      i != callback_list.end(); i++)
    delete *i;
  callback_list.clear();

  /*
   * Reset cx/cy.
   */
  cx = (this->w() - 3 * 52) / 4;
  cy = 4;
}

void LaunchBrowser::add_button(LB_Action *a) {
  string label = a->label();
  char *icon_name = strdup(label.c_str());
  Fl_Pixmap *pix = a->large_pixmap();

  begin();
  Fl_Icon* i = new Fl_Icon(cx, cy, 52, 52, icon_name, pix);
  i->labelsize(labelsize());
  i->box(FL_RFLAT_BOX);
  i->color(color());
  i->selection_color(selection_color());
  i->callback(action_cb, (void*)a);
  end();
  icon_list.push_back(i);
  if(pix)
    pixmap_list.push_back(pix);
  callback_list.push_back(a);
	
  cx += 52 + (w() - 3 * 52) / 4;
  if(cx > w() - 52) {
    cx = (w() - 3 * 52) / 4;
    cy += 52;
  }
}

void LaunchBrowser::action_cb(Fl_Widget *widget, void *data) {
  LB_Action *a = reinterpret_cast<LB_Action*>(data);
  LaunchBrowser *lb = a->launchbrowser();
  lb->do_sound();
  a->action();
}

void LaunchBrowser::category(const string &name) {
//current_folder = string(name, 0, 9);
  current_folder = string(name, 0, 8);
  launchpad_label->label(current_folder.c_str());
  launchpad_label->clear_output();
  launchpad_label->redraw();
//   if(action_stack.size() > 1) {
//     // back_button->activate();
//     home_button->activate();
//   } else {
//     // back_button->deactivate();
//     home_button->deactivate();
//   }
}

void LaunchBrowser::do_sound(void) const {
  if(!sound)
    return;
  XKeyboardControl vals;
  vals.bell_pitch = 440;
  vals.bell_duration = 5;
  XChangeKeyboardControl(fl_display, KBBellPitch | KBBellDuration, &vals);
  XBell(fl_display, 100);
  XFlush(fl_display);
}

int LaunchBrowser::handle(int event) {
  switch(event) {
  case FL_FOCUS:
  case FL_ENTER:
  case FL_SHOW: {
    if(refreshing)
      return 0;
    static time_t pt = 0;
    time_t t = time(0);
    if(t - pt > 2 && current_folder == string("Active")) {
      pt = t;
      refresh();
      return 0;
    }
  }
  }
  return Fl_Paged_Scroll::handle(event);
}
