// XActiveWindows.cxx
//
// Copyright (C) 2001 Brian Webb (webbb@desertscenes.net)
// (C) 2001-2004 Brian Webb, Andreas Holzer, Enrique Vidal
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "XActiveWindow.h"

#include <algo.h>
#include <X11/Xutil.h>
#include <X11/Xmu/WinUtil.h>
#include <stdio.h>

using std::string;

XActiveWindowGen::XActiveWindowGen(Display *_display, int screen)
  : cur_window(0), display(_display) {
  Window root = XRootWindow(display, screen);
  Window root_root;
  Window root_parent;
  Window *rootch;

  /*
   * Recreate the list of active windows.
   */
  XQueryTree(display, root, &root_root, &root_parent, &rootch,
	     &nwindows);
  if(nwindows > 0) {
    root_children = new Window[nwindows];
    copy(rootch, rootch + nwindows, root_children);
    XFree(rootch);
  }

  /*
   * Find the first valid window.
   */
  next();
}

XActiveWindowGen::XActiveWindowGen(const XActiveWindowGen &aw)
  : nwindows(aw.nwindows), cur_window(aw.cur_window), display(aw.display) {
  root_children = new Window[aw.nwindows];
  copy(aw.root_children, aw.root_children + nwindows, root_children);
}

static bool get_name(Display *display, Window wid, string &name) {
  XTextProperty text;
  char** text_list;
  int nsb;
  XGetWMName(display, wid, &text);
  if(XTextPropertyToStringList(&text, &text_list, &nsb)) {
    if(nsb > 0) {
      name = string(text_list[0]);
      XFreeStringList(text_list);
      return true;
    }
  }
  return false;
}

static bool get_command(Display *display, Window wid, string &cmd) {
  char **argv;
  int argc;
  if(XGetCommand(display, wid, &argv, &argc) != 0 && argc > 0) {
    cmd = string(argv[0]);
    XFreeStringList(argv);
    return true;
  }
  return false;
}

static bool check_client(Display *display, Window xwindow, Atom state_atom) {
  bool valid_client = true;

  string wname;
  get_name(display, xwindow, wname);

  Atom dummy1;
  int format = 0;
  unsigned long dummy2, nitems = 0, *prop = NULL;
  XGetWindowProperty(display, xwindow, state_atom, 0, 1024, False,
		     state_atom, &dummy1, &format, &nitems, &dummy2,
		     (unsigned char **)&prop);
  if(prop) {
    valid_client = format == 32 && nitems > 0 &&
      (prop[0] == NormalState || prop[0] == IconicState);
    XFree (prop);
  } else
    valid_client = false;

  /*
   * Check that the window is viewable.
   */
  if(valid_client) {
    XWindowAttributes attributes = { 0, };

    XGetWindowAttributes (display, xwindow, &attributes);
    valid_client = (attributes.c_class == InputOutput &&
		    attributes.map_state == IsViewable);
  }

  /*
   * See if the window is requesting focus.
   */
  if(valid_client) {
    XWMHints *hints = XGetWMHints (display, xwindow);

    valid_client &= !hints || !(hints->flags & InputHint) || hints->input;
    if(hints)
      XFree(hints);
  }

  return valid_client;
}

static Window find_input_client(Display *display, Window xwindow,
				Atom state_atom) {
  Window dummy1, dummy2, *children = NULL;
  unsigned int n_children = 0;
  unsigned int i;

  /*
   * Is this window an appropriate window?  If so, return it.
   */
  if(check_client(display, xwindow, state_atom))
    return xwindow;

  /*
   * Check the window's children to see if there are any appropriate windows
   * there.
   */
  if(!XQueryTree(display, xwindow, &dummy1, &dummy2, &children, &n_children) ||
     !children)
    return None;
  for(i = 0; i < n_children; i++) {
    xwindow = find_input_client(display, children[i], state_atom);
    if(xwindow)
      break;
  }
  
  XFree(children);
  
  return xwindow;
}

void XActiveWindowGen::next() {

  cur.name = string("");
  cur.xwindowid = 0;

  while(!done()) {
    Atom a = XInternAtom(display, "WM_STATE", False);
    Window wid = 0;

    if((wid = find_input_client(display, root_children[cur_window++], a)))
      if(get_name(display, wid, cur.name)) {
	get_command(display, wid, cur.command);
	cur.xwindowid = wid;
	break;
      }
  }
}
