// launchpad.cxx
//
// Copyright (C) 2000 Agenda Computing Inc.
// Written by Derek K.T. Au <phantom@netcom.ca>
// (C) 2001-2004 Brian Webb, Andreas Holzer, Enrique Vidal
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include <sys/time.h>
#include <sys/types.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <stdio.h>
#include <map>
#include <string>

#include <FL/Fl.H>
#include <FL/Fl_Box.H>
#include <FL/filename.H>
#include <Fl/Fl_Output.h>

#include <Flek/Fl_App_Window.H>

#include <flpda/Widget_Factory.h>
#include <flpda/gettext.h>
#include <flpda/pref.h>
#include <runone.h>
#ifdef VRSM
#include <LaunchPad.h>
#endif

#include "flpad.h"
#include "on_screen.h"
#include "myabout.h"
#include "Fl_XPM.h"
#include "LaunchBrowser.h"
#include "BatteryStatus.h"

using std::string;

// #define APPNAME _("Launch Pad") // Defined in flpad.h
// #define VERSION 1               // Defined in Makefile
// #define PATCHLEVEL 7            // Defined in akefile

#ifdef OLD_UPDT  // Old method based on reading /dev/vrbuttons
// For Automatic updating date/time and batt meter by power-button or pen-out
#include <fcntl.h>
#include <linux/buttons.h>
int fbtn; // file descriptor for dev/vrbuttons
#endif

Fl_Button *dateTime_label;
Fl_Button *launchpad_label;
BatteryStatus *battery_status;

static char time_string[256];
static char pref_path[1024];
static std::map<string, string> *pref_data;

static          int  sound = DEF_SOUND;              // sound enabled
static          int  battType = DEF_BATT_TYPE;       // NiMH
static          char *dateTimeFmt = DEF_TIME_FORMAT; // 24h time, day, dayofweek
static unsigned int  timePeriod = 1000*DEF_TIME_REFRESH; // refresh every 20 s.
static          int  battTimePeriod = int(DEF_BATT_REFRESH / DEF_TIME_REFRESH);
          // default batt period = 200s: should be > apm_sleep auto-off time !!!

static void about_cb(Fl_Widget*, void*) {
  char version[64];
  sprintf(version, "%d.%d", VERSION, PATCHLEVEL);
  new myAbout(_(APPNAME), version);
}

void done_cb(Fl_Widget* widget, void* data) {
  Fl_Window* window = widget->window();
  while((window = window->window())->parent() != 0);
  window->hide();
}

static void sound_cb(Fl_Widget *widget, void* data) {
  LaunchBrowser *lb = reinterpret_cast<LaunchBrowser*>(data);
  Fl_Menu_* mw = (Fl_Menu_*)widget;
  const Fl_Menu_Item* m = mw->mvalue();
  lb->sound = (m->value() != 0); sound = lb->sound;
#ifdef VRSM
  LaunchPad::sound = lb->sound;
#endif
  if(lb->sound)
    (*pref_data)[string("sound")] = string("1");
  else
    (*pref_data)[string("sound")] = string("0");
  preference_write(pref_path, pref_data);
}
void back_cb(Fl_Widget* widget, void* data) {
  LaunchBrowser* l = (LaunchBrowser*)data;
  l->back();
}

void home_cb(Fl_Widget* widget, void* data) {
  LaunchBrowser* l = (LaunchBrowser*)data;
  l->home();
}

void arg_cb(int c, char** v, void* data) {
  Fl_App_Window* l = (Fl_App_Window*)data;
  l->show();
}

static void set_current_time(void) {

  struct timeval tv;
  gettimeofday(&tv, 0);

  struct tm* display_time_tm = localtime(&tv.tv_sec);
  strftime(time_string, 19, dateTimeFmt,  display_time_tm); 
  dateTime_label->clear_output();  dateTime_label->redraw();
}

static void update_timeAndBatt(void *data) {
  set_current_time();
  static int bcount = 0;
//fprintf(stderr, "%d-", bcount);
  if (++bcount == battTimePeriod) // Should be > apm-sleep auto-off time !!!
    {battery_status->set(); bcount = 0;} 
  Fl::add_timeout(timePeriod, update_timeAndBatt, data);
}

void exec_cb(Fl_Widget* widget, void* data) {
   char tools_path[1024] = "";
   int menuCall    = (((Fl_Menu_*)widget)->mvalue() != NULL);
   int eventIsDrag = ! (Fl::event_is_click() || menuCall);
   int clicks      = Fl::event_clicks(); 

   set_current_time(); battery_status->set();

   if (eventIsDrag) filename_expand(tools_path, ((char **)data)[1]);
   else             filename_expand(tools_path, ((char **)data)[0]);
// fprintf(stderr, "%d %d %s\n", menuCall, clicks, tools_path);

   if (strlen(tools_path) > 0) {
      if (clicks == 0) on_screen(tools_path);
      char *argv[] = { (char*)tools_path, 0 };
      runone(argv);
   }
   Fl::event_clicks(0);
}

#ifdef OLD_UPDT  // Old method based on reading /dev/vrbuttons
void vrbuttons_cb(int, void*) { // Automatic updating by power-button or pen-out
  unsigned short buff;
  usleep(10000);
  if (read(fbtn, &buff, 2) != 2) 
     fprintf(stderr, "Error reading dev/vrbuttons\n");
  if ( ( ((buff & BTN_STATE_MASK) == BTN_PRESS)   &&
         ((buff & BTN_DATA_MASK)  == BTN_PEN ) )      ||
       ( ((buff & BTN_STATE_MASK) == BTN_RELEASE) && 
         ((buff & BTN_DATA_MASK)  == BTN_POWER_GPIO) ) )
    {set_current_time(); battery_status->set();}
//fprintf(stderr, "Event code=0x%04x\n", buff);
}
#else
void sigCont_cb(int) { // Automatic updating by wake-up event (SIGCONT signal)
//fprintf(stderr, "SIGCONT signal arrived\n");
  set_current_time(); battery_status->set();
}
#endif

int main(int argc, char* argv[]) {

  char *argv_batt[]= {DEF_BATT_TOOL,  DEF_BATT_TOOL1};
  char *argv_date[]= {DEF_TIME_TOOL,  DEF_TIME_TOOL1};
  char *argv_tool[]= {DEF_LABEL_TOOL, DEF_LABEL_TOOL1};

  signal(SIGCHLD, SIG_IGN);


#ifdef OLD_UPDT  // Old method based on reading /dev/vrbuttons
  // For the automatic updating by power button or pen out
  if ((fbtn = open("/dev/vrbuttons", O_RDONLY)) < 0) 
     fprintf(stderr, "Error %d opening /dev/vrbuttons\n", fbtn);
  Fl::add_fd(fbtn, vrbuttons_cb);
#else
// for asyncronous updating of date/time and battery status by wake-up event
   if (signal(SIGCONT, sigCont_cb) == SIG_ERR)
      perror("Installing SIGCONT handler"); 
#endif

  // reading preferences file
  filename_expand(pref_path, "~/.launchpad");
  pref_data = new std::map<string, string>;
  preference_read(pref_path, pref_data);
  std::map<string, string>::iterator itr;
  itr = pref_data->find(string(SOUND_STR));
  if (itr != pref_data->end()) sound = atoi((itr->second).c_str());
  itr = pref_data->find(string(BATT_TYPE_STR));
  if (itr != pref_data->end()) battType = atoi((itr->second).c_str());
  itr = pref_data->find(string(TIME_REFRESH_STR));
  if (itr != pref_data->end()) timePeriod = 1000 * atoi((itr->second).c_str());
  itr = pref_data->find(string(BATT_REFRESH_STR));
  if (itr != pref_data->end())
     battTimePeriod = int(1000 * atoi((itr->second).c_str()) / timePeriod);

  itr = pref_data->find(string(TIME_FORMAT_STR));
  if (itr != pref_data->end()) dateTimeFmt = strdup((itr->second).c_str());

  itr = pref_data->find(string(BATT_TOOL_STR));
  if (itr != pref_data->end()) argv_batt[0] = strdup((itr->second).c_str());
  itr = pref_data->find(string(BATT_TOOL1_STR));
  if (itr != pref_data->end()) argv_batt[1] = strdup((itr->second).c_str());
  itr = pref_data->find(string(TIME_TOOL_STR));
  if (itr != pref_data->end()) argv_date[0] = strdup((itr->second).c_str());
  itr = pref_data->find(string(TIME_TOOL1_STR));
  if (itr != pref_data->end()) argv_date[1] = strdup((itr->second).c_str());
  itr = pref_data->find(string(LABEL_TOOL_STR));
  if (itr != pref_data->end()) argv_tool[0] = strdup((itr->second).c_str());
  itr = pref_data->find(string(LABEL_TOOL1_STR));
  if (itr != pref_data->end()) argv_tool[1] = strdup((itr->second).c_str());

//fprintf(stderr, "Prefs: %d, %d, %d, %s\n",
//                 sound, timePeriod, battTimePeriod, dateTimeFmt);

#ifdef VRSM
  new LaunchPad;
  FILE *fp = fopen("/tmp/vrsm.pid", "w");
  if(fp) {
//fprintf(fp, "%d\n", getpid());
    fclose(fp);
  } else {
    fprintf(stderr, "Error opening /tmp/vrsm.pid.");
  }
#endif

  Fl_App_Window* window = WidgetFactory::new_window("Launch Pad");

  textdomain("flutil");

  LaunchBrowser* launch_browser =
    new LaunchBrowser(0, 0, window->w(), window->h());
  launch_browser->box(FL_FLAT_BOX);
  launch_browser->color(FL_WHITE);
  launch_browser->selection_color(FL_BLACK);
  launch_browser->labelsize(WidgetFactory::labelsize());
  launch_browser->type(0);
  launch_browser->window()->resizable(launch_browser);
  launch_browser->auto_adjust();

  window->end();

  Fl_Dockable_Window* toolbar = WidgetFactory::new_toolbar();
  unsigned int cur_width = 0;

  /*
   * Create the menu.
   */
  Fl_Menu_Button* menu = WidgetFactory::new_menu_button("Menu");
  menu->box(FL_FLAT_BOX);
  menu->resize(menu->x(), menu->y(), 28, menu->h());
  menu->add(_("About"), 0, about_cb, 0);

  int sound_flags = FL_MENU_TOGGLE;
  std::map<string, string>::iterator sitr =
    pref_data->find(string("sound"));
  if(sitr != pref_data->end() && sitr->second != string("0")) {
    sound_flags |= FL_MENU_VALUE;
    launch_browser->sound = true;
#ifdef VRSM
    LaunchPad::sound = true;
#endif
  }
#ifdef VRSM
  else
    LaunchPad::sound = false;
#endif
  menu->add(_("Sound"), 0, sound_cb, launch_browser, sound_flags);

#ifndef VRSM
  menu->add(_("Done"), 0, done_cb, 0);
#endif
  cur_width += menu->w();

  /*
   * Create the navigation buttons.
   */
//  back_button =
//    WidgetFactory::new_button("@<", back_cb, (void*)launch_browser);
//  back_button->labeltype(FL_SYMBOL_LABEL);
//  back_button->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
//  back_button->size(0, back_button->h());
//  cur_width += back_button->w();

//    home_button->size(0, home_button->h());
//  home_button =
//  WidgetFactory::new_button("@2<", home_cb, (void*)launch_browser);
//  home_button->labeltype(FL_SYMBOL_LABEL);
//  home_button->align(FL_ALIGN_CENTER|FL_ALIGN_INSIDE);
//  home_button->size(0, home_button->h());
////home_button->size(18, home_button->h());
//  cur_width += home_button->w();

  /* A thiny separator. */
  Fl_Box *sep1 = new Fl_Box(menu->x() + menu->w(),
                               menu->y(), 1,
                               menu->h(), "");
  sep1->box(FL_THIN_UP_BOX);
  cur_width += sep1->w();

  /*
   * Create the battery status indicator.
   */
  battery_status = new BatteryStatus(menu->x() + menu->w() + 1,
				     menu->y(), 12, menu->h(), battType, 0);
  battery_status->box(FL_FLAT_BOX);
  battery_status->callback(exec_cb, (void *) argv_batt );
  cur_width += battery_status->w();

  /*
   * Create the date/time label.
   */
  dateTime_label = new Fl_Button(battery_status->x() + battery_status->w(),
                                 battery_status->y(), 67,
		                 battery_status->h(), "--:-- --- --");
  dateTime_label->labelsize(10);
  dateTime_label->labelcolor(FL_BLACK);
  dateTime_label->redraw();
  dateTime_label->labelfont(FL_HELVETICA_BOLD);
  dateTime_label->box(FL_FLAT_BOX);
  dateTime_label->label(time_string);
  dateTime_label->callback(exec_cb, (void *) argv_date);

  cur_width += dateTime_label->w();

  /* A thiny separator. */
  Fl_Box *sep2 = new Fl_Box(dateTime_label->x() + dateTime_label->w(),
                            dateTime_label->y(), 1,
                            dateTime_label->h(), "");
  sep2->box(FL_THIN_UP_BOX);
  cur_width += sep2->w();

  /*
   * Create the category label.
   */
  launchpad_label = new Fl_Button(dateTime_label->x() + dateTime_label->w() + 1,
                                  dateTime_label->y(), 56,
                                  dateTime_label->h(), ROOTPAD_NAME);
  launchpad_label->labelsize(10);
  launchpad_label->labelcolor(FL_BLACK);
  launchpad_label->redraw();
  launchpad_label->box(FL_FLAT_BOX);
  // launchpad_label->box(FL_THIN_UP_BOX);
  launchpad_label->callback(exec_cb, (void *) argv_tool);

  /*
   * Create the scroll buttons.
   */
  Fl_Page_UpDown_Button* page_button = WidgetFactory::new_pagekey();
  launch_browser->page_updown_button(page_button);
  cur_width += page_button->w();

  /*
   * Size the label to fill out the toolbar.
   */
  launchpad_label->size(toolbar->w() - cur_width, menu->h());

  toolbar->end();
  window->add_dockable(toolbar, 1);

  launch_browser->refresh();
  update_timeAndBatt(launch_browser);

  string argv0;
  if(argv[0][0] != '/')
    argv0 = string("/usr/bin/") + string(argv[0]);
  else
    argv0 = string(argv[0]);
  argv[0] = (char*)argv0.c_str();
//printf("argv0: %s\n", argv[0]);
  runone_callback(argc, argv, arg_cb, (void*)window);

  Fl::run();
}
